# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Rotation matrix utilities."""

from __future__ import annotations

import logging

import numpy as np

from weitsicht.utils import Array3x3, Vector3D

__all__ = ["Rotation"]

logger = logging.getLogger(__name__)


class Rotation:
    """Represent a 3×3 rotation matrix and convert to/from common photogrammetry angle notations.

    Supported angle sets:

    - **OPK**: omega/phi/kappa (radians/degrees)
    - **APK**: alpha/zeta/kappa (radians/degrees)

    Axis conventions in weitsicht:
    - X axis: left → right in the image
    - Y axis: bottom → top in the image
    - Z axis: backwards (away from the scene)
    - Global system: ENU
    """

    def __init__(self, rotation_matrix: np.ndarray) -> None:
        """Create a :class:`Rotation` from a rotation matrix.

        :param rotation_matrix: Rotation matrix (3×3).
        :type rotation_matrix: numpy.ndarray
        """

        self._rotation_matrix = rotation_matrix

    @classmethod
    def from_opk_degree(cls, omega: float, phi: float, kappa: float) -> Rotation:
        """Create a rotation from omega/phi/kappa angles in degrees.

        :param omega: Omega angle in degrees.
        :type omega: float
        :param phi: Phi angle in degrees.
        :type phi: float
        :param kappa: Kappa angle in degrees.
        :type kappa: float
        :return: Rotation instance.
        :rtype: Rotation
        """

        rotation = cls(np.eye(3, 3, dtype=float))
        rotation.opk_degree = np.array([omega, phi, kappa])
        return rotation

    @classmethod
    def from_opk(cls, omega: float, phi: float, kappa: float) -> Rotation:
        """Create a rotation from omega/phi/kappa angles in radians.

        :param omega: Omega angle in radians.
        :type omega: float
        :param phi: Phi angle in radians.
        :type phi: float
        :param kappa: Kappa angle in radians.
        :type kappa: float
        :return: Rotation instance.
        :rtype: Rotation
        """

        rotation = cls(np.eye(3, 3, dtype=float))
        rotation.opk = np.array([omega, phi, kappa])
        return rotation

    @classmethod
    def from_apk_degree(cls, alpha: float, zeta: float, kappa: float) -> Rotation:
        """Create a rotation from alpha/zeta/kappa angles in degrees.

        ``alpha/zeta/kappa`` (often also written AZK) is a common photogrammetry notation (also used in monoplotting)
        where:

        - ``alpha``: azimuth of the camera +Z axis, relative to +X (East) in the ground CRS,
        - ``zeta``: inclination/tilt relative to nadir (0° = nadir),
        - ``kappa``: rotation around the camera +Z axis.

        :param alpha: Alpha angle in degrees.
        :type alpha: float
        :param zeta: Zeta angle in degrees.
        :type zeta: float
        :param kappa: Kappa angle in degrees.
        :type kappa: float
        :return: Rotation instance.
        :rtype: Rotation
        """

        rotation = cls(np.eye(3, 3, dtype=float))
        rotation.apk_degree = np.array([alpha, zeta, kappa])
        return rotation

    @classmethod
    def from_apk(cls, alpha: float, zeta: float, kappa: float) -> Rotation:
        """Create a rotation from alpha/zeta/kappa angles in radians.

        See :meth:`from_apk_degree` for the meaning of the angles.

        :param alpha: Alpha angle in radians.
        :type alpha: float
        :param zeta: Zeta angle in radians.
        :type zeta: float
        :param kappa: Kappa angle in radians.
        :type kappa: float
        :return: Rotation instance.
        :rtype: Rotation
        """

        rotation = cls(np.eye(3, 3, dtype=float))
        rotation.apk = np.array([alpha, zeta, kappa])
        return rotation

    @property
    def matrix(self) -> Array3x3:
        """Return the rotation matrix.

        :return: Rotation matrix (3×3).
        :rtype: Array3x3
        """
        return self._rotation_matrix

    @matrix.setter
    def matrix(self, rot: Array3x3):
        """Set the rotation matrix.

        :param rot: Rotation matrix (3×3).
        :type rot: Array3x3
        """
        self._rotation_matrix = rot

    @property
    def opk(self) -> Vector3D:
        """Convert the rotation matrix to omega/phi/kappa angles in radians.

        :return: ``[omega, phi, kappa]`` in radians.
        :rtype: Vector3D
        """

        omega_rad = np.arctan2(-self._rotation_matrix[1, 2], self._rotation_matrix[2, 2])

        # avoid a np domain error: argument of arcsin must be within [-1,1]
        phi_rad = np.arcsin(max(-1.0, min(1.0, self._rotation_matrix[0, 2])))
        kappa_rad = np.arctan2(-self._rotation_matrix[0, 1], self._rotation_matrix[0, 0])
        op_vec_rad = np.array([omega_rad, phi_rad, kappa_rad])

        return op_vec_rad

    @opk.setter
    def opk(self, opk_vec_rad: Vector3D):
        """Convert omega/phi/kappa angles (radians) to a rotation matrix.

        :param opk_vec_rad: ``[omega, phi, kappa]`` in radians.
        :type opk_vec_rad: Vector3D
        :raises ValueError: If the input vector does not have shape ``(3,)``.
        """
        if opk_vec_rad.shape == (3,):  # convert to rotation matrix
            omega = opk_vec_rad[0]
            phi = opk_vec_rad[1]
            kappa = opk_vec_rad[2]

            sin_omega = np.sin(omega)
            cos_omega = np.cos(omega)
            sin_phi = np.sin(phi)
            cos_phi = np.cos(phi)
            sin_kappa = np.sin(kappa)
            cos_kappa = np.cos(kappa)

            # This is R, not R.T !
            opk_rot_mat = np.array(
                [
                    [cos_phi * cos_kappa, -cos_phi * sin_kappa, sin_phi],
                    [
                        cos_omega * sin_kappa + sin_omega * sin_phi * cos_kappa,
                        cos_omega * cos_kappa - sin_omega * sin_phi * sin_kappa,
                        -sin_omega * cos_phi,
                    ],
                    [
                        sin_omega * sin_kappa - cos_omega * sin_phi * cos_kappa,
                        sin_omega * cos_kappa + cos_omega * sin_phi * sin_kappa,
                        cos_omega * cos_phi,
                    ],
                ]
            )

            self._rotation_matrix = opk_rot_mat

        else:
            raise ValueError(f"Vector3D expected, shape is: {opk_vec_rad.shape}")

    @property
    def opk_degree(self) -> Vector3D:
        """Return omega/phi/kappa angles in degrees.

        :return: ``[omega, phi, kappa]`` in degrees.
        :rtype: Vector3D
        """

        return self.opk / np.pi * 180

    @opk_degree.setter
    def opk_degree(self, opk_vec_degree: Vector3D):
        """Convert omega/phi/kappa angles (degrees) to a rotation matrix.

        :param opk_vec_degree: ``[omega, phi, kappa]`` in degrees.
        :type opk_vec_degree: Vector3D
        :raises ValueError: If the input vector does not have shape ``(3,)``.
        """
        if opk_vec_degree.shape == (3,):
            self.opk = opk_vec_degree * np.pi / 180
        else:
            raise ValueError(f"Vector3D expected, shape is: {opk_vec_degree.shape}")

    @property
    def apk(self) -> Vector3D:
        """Convert the rotation matrix to alpha/zeta/kappa angles in radians.

        In ``weitsicht`` the camera CRS uses a backwards Z axis. Therefore this property uses a ZYZ Euler
        decomposition of the camera's **+Z axis**:

        - ``alpha``: azimuth of the camera +Z axis (XY plane),
        - ``zeta``: off-nadir angle (0 = nadir, 90° = horizontal),
        - ``kappa``: rotation around the camera +Z axis.

        :return: ``[alpha, zeta, kappa]`` in radians.
        :rtype: Vector3D
        """

        # avoid a np domain error: argument of arccos must be within [-1,1]
        cos_zeta = max(-1.0, min(1.0, float(self._rotation_matrix[2, 2])))
        zeta_rad = float(np.arccos(cos_zeta))

        sin_zeta = float(np.sin(zeta_rad))
        if abs(sin_zeta) < 1e-12:
            # Gimbal lock: alpha and kappa are not uniquely defined for zeta ~ 0 or pi.
            # Convention here: set alpha=0 and fold everything into kappa.
            alpha_internal = 0.0
            kappa_internal = float(np.arctan2(self._rotation_matrix[1, 0], self._rotation_matrix[0, 0]))
        else:
            alpha_internal = float(np.arctan2(self._rotation_matrix[1, 2], self._rotation_matrix[0, 2]))
            kappa_internal = float(np.arctan2(self._rotation_matrix[2, 1], -self._rotation_matrix[2, 0]))

        return np.array([alpha_internal, zeta_rad, kappa_internal])

    @apk.setter
    def apk(self, apk_vec_rad: Vector3D):
        """Convert alpha/zeta/kappa angles (radians) to a rotation matrix.

        :param apk_vec_rad: ``[alpha, zeta, kappa]`` in radians.
        :type apk_vec_rad: Vector3D
        :raises ValueError: If the input vector does not have shape ``(3,)``.
        """
        if apk_vec_rad.shape == (3,):
            alpha = float(apk_vec_rad[0])
            zeta = float(apk_vec_rad[1])
            kappa = float(apk_vec_rad[2])

            sin_alpha = np.sin(alpha)
            cos_alpha = np.cos(alpha)
            sin_zeta = np.sin(zeta)
            cos_zeta = np.cos(zeta)
            sin_kappa = np.sin(kappa)
            cos_kappa = np.cos(kappa)

            # This is R, not R.T !
            # R = Rz(alpha) * Ry(zeta) * Rz(kappa)
            apk_rot_mat = np.array(
                [
                    [
                        cos_alpha * cos_zeta * cos_kappa - sin_alpha * sin_kappa,
                        -cos_alpha * cos_zeta * sin_kappa - sin_alpha * cos_kappa,
                        cos_alpha * sin_zeta,
                    ],
                    [
                        sin_alpha * cos_zeta * cos_kappa + cos_alpha * sin_kappa,
                        -sin_alpha * cos_zeta * sin_kappa + cos_alpha * cos_kappa,
                        sin_alpha * sin_zeta,
                    ],
                    [-sin_zeta * cos_kappa, sin_zeta * sin_kappa, cos_zeta],
                ]
            )

            self._rotation_matrix = apk_rot_mat
        else:
            raise ValueError(f"Vector3D expected, shape is: {apk_vec_rad.shape}")

    @property
    def apk_degree(self) -> Vector3D:
        """Return alpha/zeta/kappa angles in degrees.

        :return: ``[alpha, zeta, kappa]`` in degrees.
        :rtype: Vector3D
        """

        return self.apk / np.pi * 180

    @apk_degree.setter
    def apk_degree(self, apk_vec_degree: Vector3D):
        """Convert alpha/zeta/kappa angles (degrees) to a rotation matrix.

        :param apk_vec_degree: ``[alpha, zeta, kappa]`` in degrees.
        :type apk_vec_degree: Vector3D
        :raises ValueError: If the input vector does not have shape ``(3,)``.
        """
        if apk_vec_degree.shape == (3,):
            self.apk = apk_vec_degree * np.pi / 180
        else:
            raise ValueError(f"Vector3D expected, shape is: {apk_vec_degree.shape}")

    def transform_to_crs(self):
        """Transform the rotation into another CRS.

        This is currently not implemented.
        """

        pass



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Utilities for intersecting lines with a regular grid."""

import numpy as np

__all__ = ["vector_projection", "raster_index_p1_p2", "line_grid_intersection_points"]


def vector_projection(p1, p2, g_1):
    """Project points onto a line defined by two points.

    :param p1: First point defining the line (x, y).
    :type p1: array-like
    :param p2: Second point defining the line (x, y).
    :type p2: array-like
    :param g_1: Points to project, shape ``(N, 2)``.
    :type g_1: array-like
    :return: Projected points, shape ``(N, 2)``.
    :rtype: numpy.ndarray
    """
    b = np.zeros(g_1.shape, dtype=float) + (p2 - p1)
    a = g_1 - p1

    return (np.einsum("ij,ij->i", a, b) / np.einsum("ij,ij->i", b, b) * b.T).T + p1


def line_grid_intersection_points_iterative(p_line_1, p_line_2):  # pragma: no cover
    """This works but need more testing and edge case checking"""
    x1 = p_line_1[0]
    y1 = p_line_1[1]
    x2 = p_line_2[0]
    y2 = p_line_2[1]

    _x = np.floor(x1) + 0.5
    _y = np.floor(y1) + 0.5

    _x_end = np.floor(x2) + 0.5
    _y_end = np.floor(y2) + 0.5

    if np.isclose(abs(x2 - x1), 0):
        _x_dir = 1
    else:
        _x_dir = np.round((x2 - x1) / abs(x2 - x1))

    if np.isclose(abs(y2 - y1), 0):
        _y_dir = 1
    else:
        _y_dir = np.round((y2 - y1) / abs(y2 - y1))

    u_vec = (p_line_2 - p_line_1) / np.linalg.norm(p_line_1 - p_line_2)

    list_index = [[_x, _y]]
    while True:
        c1 = np.array([_x + _x_dir, _y])
        c2 = np.array([_x, _y + _y_dir])

        _d_c1 = np.linalg.norm(np.cross((c1 - p_line_1), u_vec))
        _d_c2 = np.linalg.norm(np.cross((c2 - p_line_1), u_vec))
        if _d_c1 < _d_c2:
            _x = _x + _x_dir

        else:
            _y = _y + _y_dir

        list_index.append([_x, _y])

        if (_x == _x_end) and (_y == _y_end):
            break

    return np.floor(np.array(list_index)).astype(int)


def raster_index_p1_p2(p_line_1, p_line_2):
    """Return the bounding raster indices covering a line segment.

    :param p_line_1: Start point (x, y).
    :type p_line_1: array-like
    :param p_line_2: End point (x, y).
    :type p_line_2: array-like
    :return: Tuple ``(low_x, high_x, low_y, high_y)`` as integer indices.
    :rtype: tuple[int, int, int, int]
    """
    x1 = p_line_1[0]
    y1 = p_line_1[1]
    x2 = p_line_2[0]
    y2 = p_line_2[1]

    if x1 <= x2:
        low_x = int(np.floor(x1))
        high_x = int(np.ceil(x2))
    else:
        high_x = int(np.ceil(x1))
        low_x = int(np.floor(x2))

    if y1 <= y2:
        low_y = int(np.floor(y1))
        high_y = int(np.ceil(y2))

    else:
        high_y = int(np.ceil(y1))
        low_y = int(np.floor(y2))

    return low_x, high_x, low_y, high_y


def line_grid_intersection_points(p_line_1, p_line_2):
    """Compute raster cell indices intersected by a line segment.

    :param p_line_1: Start point (x, y).
    :type p_line_1: array-like
    :param p_line_2: End point (x, y).
    :type p_line_2: array-like
    :return: Array of integer indices with shape ``(N, 2)``.
    :rtype: numpy.ndarray
    """
    # Suffers a little from floating point
    # Would need to check epsilon before doing floor
    # Should then include both cells just for safety

    x1 = p_line_1[0]
    y1 = p_line_1[1]
    x2 = p_line_2[0]
    y2 = p_line_2[1]

    if x1 <= x2:
        step_x = 1
        start_x = np.floor(x1)
        stop_x = np.ceil(x2) + 1
    else:
        step_x = -1
        start_x = np.ceil(x1)
        stop_x = np.floor(x2) - 1

    if y1 <= y2:
        step_y = 1
        start_y = np.floor(y1)
        stop_y = np.ceil(y2) + 1

    else:
        step_y = -1
        start_y = np.ceil(y1)
        stop_y = np.floor(y2) - 1

    # 1st vertical rows

    x3_matrix = np.arange(start_x, stop_x, step_x)
    x4_matrix = x3_matrix * 1.0
    # _size = np.ceil(x2) - np.floor(y1) + 1
    y3_matrix = np.ones(x3_matrix.shape) * start_y
    y4_matrix = np.ones(x3_matrix.shape) * stop_y

    p_x = np.array([]).reshape(0, 2)
    _d = (x1 - x2) * (y3_matrix - y4_matrix) - (y1 - y2) * (x3_matrix - x4_matrix)
    if not np.any(_d == 0.0):
        _t_s = (x1 - x3_matrix) * (y3_matrix - y4_matrix) - (y1 - y3_matrix) * (x3_matrix - x4_matrix)

        _u_s = (x1 - x2) * (y1 - y3_matrix) - (y1 - y2) * (x1 - x3_matrix)

        t = _t_s / _d
        _u = -_u_s / _d

        p_x = np.array([x3_matrix, y1 + t * (y2 - y1)]).T

        # If it's going to the left we need to switch the x coordinates otherwise we will miss cells
        if y1 > y2:
            p_x[:, 0] -= 0.1

    # 2nd vertical rows

    y3_matrix = np.arange(start_y, stop_y, step_y)
    y4_matrix = y3_matrix * 1.0
    x3_matrix = np.ones(y3_matrix.shape) * start_x
    x4_matrix = np.ones(y3_matrix.shape) * stop_x

    p_y = np.array([]).reshape(0, 2)
    _d = (x1 - x2) * (y3_matrix - y4_matrix) - (y1 - y2) * (x3_matrix - x4_matrix)
    if not np.any(_d == 0.0):
        _t_s = (x1 - x3_matrix) * (y3_matrix - y4_matrix) - (y1 - y3_matrix) * (x3_matrix - x4_matrix)
        _u_s = (x1 - x2) * (y1 - y3_matrix) - (y1 - y2) * (x1 - x3_matrix)

        t_vertical = _t_s / _d
        _u_vertical = -_u_s / _d

        p_y = np.array([x1 + t_vertical * (x2 - x1), y3_matrix]).T

        if x1 > x2:
            p_y[:, 1] -= 0.1

    p_all = np.vstack((p_x, p_y))

    # start and endpoint are within one cell
    if len(p_all) == 0:
        p_all = np.array([[x1, y1]])
        return np.floor(p_all).astype(int)

    _dist = np.linalg.norm(p_all - np.array([x1, y1]), axis=1)
    idx_sort_deist = np.argsort(_dist)

    p_all = p_all[idx_sort_deist, :]

    return np.floor(p_all).astype(int)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Batch operations for mapping and projecting across multiple images."""

import logging
from typing import overload

import numpy as np
from pyproj import CRS

from weitsicht.exceptions import (
    CoordinateTransformationError,
    CRSInputError,
    MapperMissingError,
    MappingError,
    NotGeoreferencedError,
)
from weitsicht.image.base_class import ImageBase
from weitsicht.mapping.base_class import MappingBase
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import ArrayNx3, Issue, MappingResult, ProjectionResult, ResultFailure, to_array_nx3

__all__ = ["ImageBatch"]

logger = logging.getLogger(__name__)


class ImageBatch:
    """Container for running operations on a set of named images."""

    def __init__(self, images: dict[str, ImageBase], mapper: MappingBase | None = None):
        """Initialize an image batch.

        :param images: Dictionary of images.
        :type images: dict[str, ImageBase]
        :param mapper: Optional default mapper assigned to images without a mapper, defaults to ``None``.
        :type mapper: MappingBase | None
        """
        self.images: dict[str, ImageBase] = images
        self.mapper: MappingBase | None = mapper

        if self.mapper is not None:
            for name, img in self.images.items():
                if img.mapper is None:
                    self.images[name].mapper = self.mapper

    def __len__(self):
        """Return the number of images in the batch."""

        return len(self.images)

    def add_images(self, images: dict[str, ImageBase]):
        """Add images to the batch.

        :param images: Images to add.
        :type images: dict[str, ImageBase]
        :raises KeyError: If any keys already exist in the batch.
        """

        equal_keys = list(set(self.images.keys()).intersection(images.keys()))

        if len(equal_keys) > 0:
            raise KeyError("Identical keys can not be added")

        if self.mapper is not None:
            for name, img in images.items():
                if img.mapper is None:
                    images[name].mapper = self.mapper

        self.images.update(images)

    def __setitem__(self, key, image: ImageBase):
        """Set an image in the batch by key.

        :param key: Image key.
        :param image: Image instance.
        :type image: ImageBase
        :raises ValueError: If the object is not an image instance.
        """

        if image.__class__.__base__ != ImageBase:
            raise ValueError("Only a image class can be assigned")

        self.images[key] = image

    @overload
    def __getitem__(self, keys: str) -> ImageBase: ...

    @overload
    def __getitem__(self, keys: tuple[str] | list[str]) -> dict[str, ImageBase]: ...

    def __getitem__(self, keys: str | tuple[str] | list[str]) -> ImageBase | dict[str, ImageBase]:
        """Get images by key or keys.

        :param keys: Image key or sequence of keys.
        :type keys: str | tuple[str] | list[str]
        :return: Single image or a dictionary of images.
        :rtype: ImageBase | dict[str, ImageBase]
        :raises KeyError: If a requested key is not found.
        :raises TypeError: If ``keys`` has an unsupported type.
        """

        if isinstance(keys, str):
            if keys in self.images.keys():
                return self.images[keys]
            else:
                raise KeyError("Key not found in images")
        elif isinstance(keys, tuple) or isinstance(keys, list):
            # raise KeyError("Key not found in images")
            # convert a simple index and tuples to lists
            # if isinstance(keys, tuple)
            # keys = list(keys) if type(keys) in [list, tuple] else list([keys])

            dict_return = {}

            for key in keys:
                if key in self.images.keys():
                    dict_return[key] = self.images[key]
                else:
                    raise KeyError("Keys not found in images")

            if len(dict_return) == 1:
                return list(dict_return.values())[0]
            return dict_return

        raise TypeError

    @overload
    def index(self, indices: int) -> ImageBase: ...

    @overload
    def index(self, indices: list[int] | tuple[int]) -> dict[str, ImageBase]: ...
    def index(self, indices: list[int] | tuple[int] | int) -> ImageBase | dict[str, ImageBase]:
        """index get images by indices

        Get the images by using indices. Will return the image at the indices of the image dict from ImageBatch.

        :param indices: The indices you want to get images for
        :type indices: list[int] | tuple[int] | int
        :raises TypeError: Indices type not valid
        :raises IndexError: One of the indices is not valid
        :return: return single image or dict of images for multiple indices
        :rtype: ImageBase | dict[str, ImageBase]
        """

        if isinstance(indices, int):
            _indices = list([indices])

        elif isinstance(indices, tuple) or isinstance(indices, list):
            _indices = list(indices)
        else:
            raise TypeError

        keys = list(self.images.keys())

        try:
            dict_return = {keys[x]: self.images[keys[x]] for x in _indices}

        except IndexError as err:
            raise IndexError("Index not found in images") from err

        if len(dict_return) == 1:
            return list(dict_return.values())[0]

        return dict_return

    def map_center_point(
        self, mapper: MappingBase | None = None, transformer: CoordinateTransformer | None = None
    ) -> dict[str, MappingResult]:
        """Map center point of image to 3d

        :param mapper: Specify Mapper to be used, can be different to the one assigned in the class
        :param transformer: A CoordinateTransfoer object which can be passed instead using crs_s
        :return: MappingResult for all images
        :raises ValueError: Image Batch is empty
        :raises NotGeoreferencedError: If an image is not geo-referenced
        :raises MapperMissingError: If no mapper is available
        :raises CRSInputError: If CRS/transformer input is invalid
        :raises CoordinateTransformationError: If coordinate transformation fails
        :raises MappingError: If mapping fails
        """

        center_p_dict = {}

        if self.__len__() == 0:
            raise ValueError("ImageBatch is empty")

        # one_mapping_worked = False
        all_mapping_worked = True
        for key, image in self.images.items():
            try:
                result = image.map_center_point(mapper=self.mapper if mapper is None else mapper, transformer=transformer)

            except NotGeoreferencedError as err_georef:
                result = ResultFailure(ok=False, error=str(err_georef), issues={Issue.IMAGE_BATCH_ERROR})
            except MapperMissingError as err_mapper:
                result = ResultFailure(ok=False, error=str(err_mapper), issues={Issue.IMAGE_BATCH_ERROR})
            except CRSInputError as err_crs:
                result = ResultFailure(ok=False, error=str(err_crs), issues={Issue.IMAGE_BATCH_ERROR})
            except CoordinateTransformationError as err_trafo:
                result = ResultFailure(ok=False, error=str(err_trafo), issues={Issue.IMAGE_BATCH_ERROR})
            except MappingError as err:
                result = ResultFailure(ok=False, error=str(err), issues={Issue.IMAGE_BATCH_ERROR})

            center_p_dict[key] = result

            if result.ok is False:
                all_mapping_worked = False
            # else:
            #    one_mapping_worked = True

        # if not one_mapping_worked:
        #    raise MappingError("For none of the images the center point could be mapped")

        if not all_mapping_worked:
            logger.warning("Some of the center points could not me mapped")

        return center_p_dict

    def map_footprint(
        self,
        points_per_edge: int = 0,
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> dict[str, MappingResult]:
        """Map footprint of images of batch to 3d

        :param points_per_edge: The number of points which should be inserted between corners.
            0: only corner points are mapped
        :param mapper: Optional specify another Mapper to be used,
            can be different to the one assigned in the image class
        :param transformer: A CoordinateTransfoer object which can be passed instead using crs_s
        :return: MappingResult for all images
        :raises ValueError: Image Batch is empty
        :raises NotGeoreferencedError: If an image is not geo-referenced
        :raises MapperMissingError: If no mapper is available
        :raises CRSInputError: If CRS/transformer input is invalid
        :raises CoordinateTransformationError: If coordinate transformation fails
        :raises MappingError: If mapping fails
        """

        footprint_dict = {}

        if self.__len__() == 0:
            raise ValueError("ImageBatch is empty")

        # one_mapping_worked = False
        all_mapping_worked = True
        for key, image in self.images.items():
            try:
                result = image.map_footprint(
                    points_per_edge=points_per_edge,
                    mapper=self.mapper if mapper is None else mapper,
                    transformer=transformer,
                )
            except NotGeoreferencedError as err_georef:
                result = ResultFailure(ok=False, error=str(err_georef), issues={Issue.IMAGE_BATCH_ERROR})
            except MapperMissingError as err_mapper:
                result = ResultFailure(ok=False, error=str(err_mapper), issues={Issue.IMAGE_BATCH_ERROR})
            except CRSInputError as err_crs:
                result = ResultFailure(ok=False, error=str(err_crs), issues={Issue.IMAGE_BATCH_ERROR})
            except CoordinateTransformationError as err_trafo:
                result = ResultFailure(ok=False, error=str(err_trafo), issues={Issue.IMAGE_BATCH_ERROR})
            except MappingError as err:
                result = ResultFailure(ok=False, error=str(err), issues={Issue.IMAGE_BATCH_ERROR})

            footprint_dict[key] = result

            if result.ok is False:
                all_mapping_worked = False
            # else:
            #    one_mapping_worked = True

        # if not one_mapping_worked:
        # raise MappingError("For none of the images a footprint could be mapped")

        if not all_mapping_worked:
            logger.warning("Some of the footprints could not me mapped")

        return footprint_dict

    def project(
        self,
        coordinates: ArrayNx3,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
        only_valid: bool = False,
    ) -> dict[str, ProjectionResult] | None:
        """Calculate projection of 3d points into image. If points was outsize image size and to_distortion is true
        the undistorted projection will be returned

        :param coordinates: Array of (nx3) with the point positions
        :type coordinates: ArrayNx3
        :param crs_s: The coordinate system of the input coordinates
        :param transformer: A CoordinateTransfoer object which can be passed instead using crs_s
        :param only_valid: If true only return images with valid projections.
        :return: dict[ImageName, ProjectionResult] or None if for filtered (only_valid) no projections are valid
        :raises ValueError: If the image batch is empty.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises NotGeoreferencedError: If an image is not geo-referenced.
        :raises MapperMissingError: If a required mapper is missing.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        if self.__len__() == 0:
            raise ValueError("ImageBatch is empty")

        _coordinates = to_array_nx3(coordinates)

        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")

        projected_dict: dict[str, ProjectionResult] = {}

        for key, image in self.images.items():
            try:
                result = image.project(_coordinates, crs_s, transformer=transformer)

            except NotGeoreferencedError as err_georef:
                result = ResultFailure(ok=False, error=str(err_georef), issues={Issue.IMAGE_BATCH_ERROR})
            except MapperMissingError as err_mapper:
                result = ResultFailure(ok=False, error=str(err_mapper), issues={Issue.IMAGE_BATCH_ERROR})
            except CRSInputError as err_crs:
                result = ResultFailure(ok=False, error=str(err_crs), issues={Issue.IMAGE_BATCH_ERROR})
            except CoordinateTransformationError as err_trafo:
                result = ResultFailure(ok=False, error=str(err_trafo), issues={Issue.IMAGE_BATCH_ERROR})

            projected_dict[key] = result

        if only_valid:
            filtered_dict: dict[str, ProjectionResult] = {}
            for key, value in projected_dict.items():
                if value.ok is True:
                    if np.any(value.mask):
                        filtered_dict[key] = value

            if len(filtered_dict) == 0:
                return None

            return filtered_dict

        return projected_dict



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

# Todo advanced database with model and make similarity search

#: Sensor (width_mm, height_mm) keyed by camera model name.
ImageSize: dict[str, tuple[float, float]]
ImageSize = {
    # ... existing entries ..
    "L1D-20c": (13.20, 8.80),
    "L2D-20c": (13.20, 8.80),
    "FC3411": (13.2, 8.8),
    "DSC-RX1RM2": (35.9, 24.00),
    "ILCE-1": (35.9, 24),
    "Canon5DSR": (36, 24),
    "GFX100S": (43.8, 32.9),
    "GFX100 II": (43.8, 32.9),
    "NIKON D3200": (23.2, 15.4),
}


def get_sensor_from_database(make, model) -> tuple | None:
    # Find sensor in database
    sensor_width = None
    if model in ImageSize:
        sensor_width = ImageSize[model]

    return sensor_width



from weitsicht.geometry.coplanar_collinear import is_coplanar


def test_is_coplanar():
    p1 = (0, 0, 0)
    p2 = (0, 100, 0)
    p3 = (100, 0, 100)

    p_test = (100, 100, 100)

    assert is_coplanar(p1, p2, p3, p_test)

    p3 = (100, 0, 100.1)

    assert not is_coplanar(p1, p2, p3, p_test)



import numpy as np

from weitsicht.geometry.coplanar_collinear import is_coplanar
from weitsicht.geometry.intersection_bilinear import multilinear_poly_intersection


def test_multilinear_poly_intersection():
    points = np.array(
        [
            [434.0, 546.0, 1022.75415039],
            [434.0, 547.0, 1024.71032715],
            [435.0, 546.0, 1022.11627197],
            [435.0, 547.0, 1022.40734863],
        ]
    )

    pos = np.array([433.34261229, 546.22363908, 1041.6013568])
    ray = np.array([0.04988232, 0.03741223, -0.99805415])

    interp = multilinear_poly_intersection(points, pos, ray)

    assert interp is not None
    assert np.allclose(interp, np.array([[434.22161617, 546.88290063, 1024.014094]]), atol=1e-8, rtol=0)

    points = np.array(
        [
            [434.0, 546.0, 1022.75415039],
            [434.0, 547.0, 1024.71032715],
            [435.0, 546.0, 1022.11627197],
            [435.0, 547.0, 1022.40734863],
        ]
    )

    pos = np.array([434.34261229, 546.22363908, 1500])
    ray = np.array([0.00, 0.0, -1])

    interp = multilinear_poly_intersection(points, pos, ray)

    # coplanar
    points = np.array(
        [
            [434.0, 546.0, 1022],
            [434.0, 547.0, 1022],
            [435.0, 546.0, 1022],
            [435.0, 547.0, 1022],
        ]
    )

    pos = np.array([434.34261229, 546.22363908, 1500])
    ray = np.array([0.00, 0.0, -1])

    interp = multilinear_poly_intersection(points, pos, ray)
    assert interp is not None
    assert np.allclose(interp, np.array([434.34261229, 546.22363908, 1022]), atol=1e-9, rtol=0)

    # Double Intersection
    points = np.array(
        [
            [434.0, 546.0, 1000],
            [434.0, 547.0, 1022],
            [435.0, 546.0, 1022],
            [435.0, 547.0, 1000],
        ]
    )

    pos = points[0, :] + np.array([0, 0, 1])
    ray = points[3, :] - points[0, :]

    interp = multilinear_poly_intersection(points, pos, ray)
    assert interp is not None
    assert np.allclose(interp[2], 1001, atol=1e-9, rtol=0)

    # No intersection
    points = np.array(
        [
            [434.0, 546.0, 1000],
            [434.0, 547.0, 1022],
            [435.0, 546.0, 1022],
            [435.0, 547.0, 1000],
        ]
    )

    pos = points[0, :] + np.array([0, 0, 100])
    ray = points[3, :] - points[0, :]

    interp = multilinear_poly_intersection(points, pos, ray)
    assert interp is None

    # Low intersection
    points = np.array(
        [
            [434.0, 546.0, 1000],
            [434.0, 547.0, 1022],
            [435.0, 546.0, 1022],
            [435.0, 547.0, 1000],
        ]
    )

    ray = points[3, :] - points[0, :]
    pos = points[0, :] - ray * 1e-7

    interp = multilinear_poly_intersection(points, pos, ray)
    assert interp is not None
    assert np.allclose(interp, points[0, :], atol=1e-9, rtol=0)

    # intersection coplanar and ray parallel
    points = np.array([[0.0, 0.0, 0], [1.0, 0.0, 0], [0.0, 1.0, 1], [1.0, 1.0, 1]])

    assert is_coplanar(*points)

    ray = points[2, :] - points[0, :]
    pos = points[0, :]

    interp = multilinear_poly_intersection(points, pos, ray)
    assert interp is None



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import logging

from weitsicht.metadata.camera_alternative_tags import AlternativeCalibrationTags
from weitsicht.metadata.camera_database import get_sensor_from_database
from weitsicht.metadata.camera_estimator_metadata import estimate_camera, ior_from_meta
from weitsicht.metadata.eor_from_meta import eor_from_meta
from weitsicht.metadata.image_from_meta import ImageFromMetaBuilder, image_from_meta
from weitsicht.metadata.tag_systems import PyExifToolTags

__all__ = [
    "AlternativeCalibrationTags",
    "get_sensor_from_database",
    "estimate_camera",
    "ior_from_meta",
    "eor_from_meta",
    "ImageFromMetaBuilder",
    "image_from_meta",
    "PyExifToolTags",
]

logger = logging.getLogger(__name__)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


"""Estimate camera intrinsics (IOR) from resolved metadata tags.

This module operates on weitsicht's tag dataclasses
(e.g. :class:`~weitsicht.metadata.tag_systems.tag_base.MetaTagIORBase`)
and is independent of how the metadata was extracted (ExifTool, PIL, custom parsers, ...).
"""

import logging
import math

from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective
from weitsicht.metadata.camera_alternative_tags import AlternativeCalibrationTags
from weitsicht.metadata.camera_database import get_sensor_from_database
from weitsicht.metadata.metadata_results import (
    IORFromMetaResult,
    IORFromMetaResultSuccess,
    MetadataIssue,
)
from weitsicht.metadata.tag_systems.tag_base import MetaTagIORBase, MetaTagIORExtended
from weitsicht.utils import ResultFailure

# Unit conversion factor
inch_to_mm = 25.4
cm_to_mm = 10
um_to_mm = 0.001

logger = logging.getLogger(__name__)


def get_unit_factor(resolution_unit) -> float | None:
    """Factor to scale the exif resolution unit to millimeter which is used for Focal length in EXIF.
    Tag 0xa20e - FocalPlaneXResolution
    Tag 0xa210 - FocalPlaneResolutionUnit
    https://www.sno.phy.queensu.ca/~phil/exiftool/TagNames/EXIF.html
    We assume square Pixels, so we only will do it for the image width side

    :param resolution_unit: the resolution unit value given in the EXIF
    :return: Unit factor or None
    """
    if resolution_unit == 2:  # Inch
        return inch_to_mm
    elif resolution_unit == 3:  # Centimeter
        return cm_to_mm
    elif resolution_unit == 4:  # Millimeter
        return 1
    elif resolution_unit == 5:  # Micrometer
        return um_to_mm
    else:
        return None


def compute_sensor_width_in_mm(image_width: int, tags: MetaTagIORBase) -> float | None:
    """Get Sensor with in mm using the image width in pixel and the tags ResolutionUnit and FocalPlaneXResolution.

    :param image_width: Width of image in Pixels
    :param tags: resolved metadata values
    :returns: Sensor width in mm or None if not possible"""

    resolution_unit = tags.focal_plane_resolution_unit
    pixels_per_unit = tags.focal_plane_x_resolution
    pixels_per_unit_y = tags.focal_plane_y_resolution

    if resolution_unit is None or pixels_per_unit is None:
        return None

    unit_factor = get_unit_factor(resolution_unit)

    if not unit_factor:
        return None

    if pixels_per_unit <= 0.0:
        # Some wrongly formatted camera have negative resolutions
        # We check if at least YResolution is present and if not negative use that
        if pixels_per_unit_y is None or pixels_per_unit_y <= 0.0:
            return None
        pixels_per_unit = pixels_per_unit_y

    pixel_pitch = unit_factor / pixels_per_unit

    return image_width * pixel_pitch


def compute_focal_length_from_35mm(
    focal_35mm: float, image_width: int | float, image_height: int | float
) -> float | None:
    # If a focal length of 35mm exists together with image width this is a good approximating
    if focal_35mm > 0:
        mm_to_pixel = 43.3 / math.sqrt(image_width**2 + image_height**2)  # 35mm film have a sensor size of 36x24mm.
        focal_pixel = focal_35mm / mm_to_pixel
        return focal_pixel
    return None


def compute_from_sensor_width(focal_mm: float, sensor_width: float, image_width: int) -> float:
    return focal_mm / (sensor_width / image_width)


def estimate_camera(tags: MetaTagIORBase) -> tuple[int, int, float, float, float]:
    """Estimate the standard camera parameters.
    Even though I am not sure if 35mm equivalent is always calculated correctly we will use that as first source,
    because for resampled images there could be that original exif-tags are left (e.g. PlaneResolution for example)
    which would be wrong for the resampled image to have correct focal length in pixel.
    Currently, 4 possible ways are implemented
    (1) Using 35mm equivalent
    (2) Using focal length and Resolution unit
    (3) Using focal length and Sensor Database; this works only if camera type is present in database
    More advanced calibration tags in XMP like that one from Pix4D will be treated separately
    The principal point is for that base estimation the center of the image

    :param meta_data: Metadata dict following exif-tools tags
    :raises ValueError: If standard Tags can not be used or if dimensions are not given
    :returns: tuple(width, height, focal length in pixel, c_x(principal point), c_y(principal point))"""

    # Get dimensions of image
    width = int(tags.image_shape[0])
    height = int(tags.image_shape[1])

    if width == 0 or height == 0:
        raise ValueError("No image dimension are present in meta data")

    # FocalLength is the real focal length in mm
    focal = tags.focal_length or 0.0

    # Focal_35mm is the focal length if the camera would be a 35mm Format camera to have the same field of view
    # If focal_35 is present it is always possible to derive focal length in pixel
    # Therefore we try this first as then we are always able to create a valid camera class
    # But as it turns out, not all vendors are very precise how this is derived compared to the focal length
    focal_35 = tags.focal_length_35mm or 0.0

    focal_pixel = None

    # 1st try - get focal length in pixel from 35mm equivalent
    if focal_35 > 0:
        focal_pixel = compute_focal_length_from_35mm(focal_35, width, height)
        logger.debug(f"Estimated Focal Length in pixel from 35mm equivalent: f={focal_pixel:6.2f}")

    # 2nd try  - estimate focal length in pixel from the tag focalLength
    # We will always try this as well even if focal_35 is present
    if focal > 0.0:
        sensor_width = compute_sensor_width_in_mm(width, tags)
        if sensor_width is not None:
            focal_pixel = compute_from_sensor_width(focal, sensor_width=sensor_width, image_width=width)
            logger.debug(f"Estimated Focal Length in pixel from sensor with: f={focal_pixel:6.2f}")

    # Principal point default will be center of image size
    # Later we try if in the metadata are better values
    c_x = width / 2.0
    c_y = height / 2.0

    # 3rd try - We check other tags from different software packages and drone vendors
    # E.g. newer DJI model save calibrated values CalibratedFocalLength, CalibratedOpticalCenterX/Y
    # Or Pix4D has specified a tag system for the IOR and also for coordinate systems
    # CalibratedFocalLength: 3666.666504
    # CalibratedOpticalCenterX: 2736.000000

    # We use a wrapper to call all different tag systems
    # res = calibration_estimator(meta_data)
    # if res is not None:
    #    width, height, focal_pixel, c_x, c_y, distortion = res

    # 4th and last try - estimate focal length from sensor database -
    # No sensor size is available in metadata
    # This is highly depending on the camera database and if the sensor is available there
    # If the sensor is not found you can add it in "camare_database.py"
    # Anyhow we will try this as this might be more accurate then estimating sensor size from exif data
    if focal > 0.0:
        sensor_size = get_sensor_from_database(make=tags.make, model=tags.model)
        if sensor_size is not None:
            focal_pixel = compute_from_sensor_width(focal, sensor_width=sensor_size[0], image_width=width)
            # logger.debug("Estimated Focal Length in pixel from sensor with from database: f=%6.2f" % focal_pixel)

    if focal_pixel is None:
        raise ValueError("Estimation of focal length in pixel failed")
    return width, height, focal_pixel, c_x, c_y


def ior_from_meta(tags_ior: MetaTagIORBase, tags_ior_extended: MetaTagIORExtended) -> IORFromMetaResult:
    """Estimate camera intrinsics (IOR) from metadata tags.

    :param tags_ior: Standard intrinsic-related tags.
    :type tags_ior: MetaTagIORBase
    :param tags_ior_extended: Optional extended calibration tags.
    :type tags_ior_extended: MetaTagIORExtended
    :return: Successful IOR result or a failure result.
    :rtype: IORFromMetaResult
    """
    try:
        width, height, focal_pixel, c_x, c_y = estimate_camera(tags=tags_ior)
    except ValueError as err:
        return ResultFailure(ok=False, error=str(err), issues={MetadataIssue.IOR_FAILED})

    # This is the model which should always work with minimum meta data present
    # At least for standard digital camera images
    # If we would start to use other camera systems like fishey models we would need to decide here which one to use
    camera = CameraOpenCVPerspective(width=width, height=height, fx=focal_pixel, fy=focal_pixel, cx=c_x, cy=c_y)

    # try to find better camera model in metadata (EXIF/XMP)
    cam_result = AlternativeCalibrationTags.run_all_methods(tags_ior=tags_ior, tags_ior_extended=tags_ior_extended)

    # It the result was not None we use that specific camera calibration, otherwise we use the standard one
    if cam_result is not None:
        camera = cam_result

    return IORFromMetaResultSuccess(ok=True, camera=camera, width=width, height=height)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import runpy
from pathlib import Path

import pytest

# Collect all top-level example scripts
EXAMPLES = sorted(Path(__file__).resolve().parent.parent.joinpath("examples").glob("*.py"))


@pytest.mark.slow
@pytest.mark.parametrize("script", EXAMPLES, ids=lambda p: p.name)
def test_example_script_runs(script, monkeypatch):
    # Ensure relative data paths in examples keep working
    monkeypatch.chdir(script.parent)
    # Execute the example as a script; any exception will fail the test
    runpy.run_path(str(script), run_name="__main__")



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Type guard helpers for narrowing weitsicht base classes.

These helpers are primarily intended for static type checkers (mypy, pyright).
They allow narrowing from ``ImageBase``/``CameraBasePerspective`` to concrete subclasses
based on the runtime type discriminator enums.
"""

from __future__ import annotations

from typing import Literal, TypeGuard, overload

from weitsicht.camera.base_perspective import CameraBasePerspective, CameraType
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective
from weitsicht.image.base_class import ImageBase, ImageType
from weitsicht.image.ortho import ImageOrtho
from weitsicht.image.perspective import ImagePerspective

__all__ = [
    "is_opencv_camera",
    "is_camera_type",
    "is_perspective_image",
    "is_ortho_image",
]


# Camera guards
def is_opencv_camera(cam: CameraBasePerspective) -> TypeGuard[CameraOpenCVPerspective]:
    """Return whether ``cam`` is an OpenCV camera.

    :param cam: Camera instance to check.
    :type cam: CameraBasePerspective
    :return: ``True`` if ``cam`` is an OpenCV-based camera.
    :rtype: TypeGuard[CameraOpenCVPerspective]
    """
    return cam.type is CameraType.OpenCV


# Overload-like behavior via Literal plus a fallback; extend when new camera types are added.
@overload
def is_camera_type(
    cam: CameraBasePerspective, kind: Literal[CameraType.OpenCV]
) -> TypeGuard[CameraOpenCVPerspective]: ...


@overload
def is_camera_type(cam: CameraBasePerspective, kind: CameraType) -> bool: ...


def is_camera_type(cam: CameraBasePerspective, kind: CameraType) -> bool:
    """Return whether ``cam.type`` matches ``kind``.

    When called with a literal discriminator, type checkers can narrow the return type
    accordingly (e.g. ``CameraType.OpenCV`` → :class:`CameraOpenCVPerspective`).

    :param cam: Camera instance to check.
    :type cam: CameraBasePerspective
    :param kind: Expected camera type discriminator.
    :type kind: CameraType
    :return: ``True`` if ``cam.type`` matches ``kind``.
    :rtype: bool
    """
    return cam.type is kind


# Image guards
def is_perspective_image(img: ImageBase) -> TypeGuard[ImagePerspective]:
    """Return whether ``img`` is a perspective image.

    :param img: Image instance to check.
    :type img: ImageBase
    :return: ``True`` if ``img`` is a perspective image.
    :rtype: TypeGuard[ImagePerspective]
    """
    return img.type is ImageType.Perspective


def is_ortho_image(img: ImageBase) -> TypeGuard[ImageOrtho]:
    """Return whether ``img`` is an orthophoto image.

    :param img: Image instance to check.
    :type img: ImageBase
    :return: ``True`` if ``img`` is an orthophoto image.
    :rtype: TypeGuard[ImageOrtho]
    """
    return img.type is ImageType.Orthophoto



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Factory helpers for creating image models from dictionaries."""

from weitsicht.image.base_class import ImageBase, ImageType
from weitsicht.image.ortho import ImageOrtho
from weitsicht.image.perspective import ImagePerspective
from weitsicht.mapping.base_class import MappingBase

__all__ = ["get_image_from_dict"]


def get_image_from_dict(param_dict: dict, mapper: MappingBase | None = None) -> ImageBase | ImageOrtho | ImagePerspective:
    """Create an image model from a parameter dictionary.

    Exceptions raised by the underlying image ``from_dict`` implementations are propagated unchanged.

    :param param_dict: Dictionary as received by :attr:`~weitsicht.image.base_class.ImageBase.param_dict`.
    :type param_dict: dict
    :param mapper: Mapping instance, defaults to ``None``.
    :type mapper: MappingBase | None
    :return: Image model instance.
    :rtype: ImageBase
    :raises KeyError: If the dictionary key ``type`` is missing or required image parameters are missing.
    :raises ValueError: If configuration values are invalid or the image type is unsupported.
    :raises TypeError: If configuration values have incompatible types.
    :raises CRSInputError: If a CRS WKT string is invalid or unsupported.
    """
    image_type = param_dict.get("type", None)
    if image_type is None:
        raise KeyError("Dictionary key 'type' is missing")

    if image_type == ImageType.Orthophoto.fullname:
        return ImageOrtho.from_dict(param_dict, mapper)

    elif image_type == ImageType.Perspective.fullname:
        return ImagePerspective.from_dict(param_dict, mapper)

    raise ValueError(f"Unsupported image type {image_type!r}")



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Factory helpers for creating camera models from dictionaries."""

from __future__ import annotations

from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.camera.camera_types import CameraType
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective

__all__ = ["get_camera_from_dict"]


def get_camera_from_dict(camera_dict: dict) -> CameraBasePerspective:
    """Create a camera model from a parameter dictionary.

    Exceptions raised by the underlying camera ``from_dict`` implementations are propagated unchanged.

    :param camera_dict: Dictionary as received by :attr:`CameraBasePerspective.param_dict`.
    :type camera_dict: dict
    :return: Camera model instance.
    :rtype: CameraBasePerspective
    :raises KeyError: If the dictionary key ``type`` is missing or required camera parameters are missing.
    :raises ValueError: If configuration values are invalid or the camera type is unsupported.
    :raises TypeError: If configuration values have incompatible types.
    """

    camera_type = camera_dict.get("type", None)
    if camera_type is None:
        raise KeyError("Dictionary key 'type' is missing")

    if camera_dict.get("type") == CameraType.OpenCV.fullname:
        return CameraOpenCVPerspective.from_dict(camera_dict)

    raise ValueError(f"Unsupported camera type {camera_type!r}")



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Coordinate transformations and rotation helpers."""

import logging

from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.transform.rotation import Rotation
from weitsicht.transform.utm_converter import get_zone, point_convert_utm_wgs84_egm2008

__all__ = [
    "CoordinateTransformer",
    "Rotation",
    "get_zone",
    "point_convert_utm_wgs84_egm2008",
]

logger = logging.getLogger(__name__)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Build images from metadata (IOR + EOR) with optional caching/tracing."""

from __future__ import annotations

from pyproj import CRS

from weitsicht.image.perspective import ImagePerspective
from weitsicht.metadata.camera_estimator_metadata import ior_from_meta
from weitsicht.metadata.eor_from_meta import eor_from_meta
from weitsicht.metadata.metadata_results import (
    EORFromMetaResult,
    ImageFromMetaResult,
    ImageFromMetaResultSuccess,
    IORFromMetaResult,
    MetadataIssue,
)
from weitsicht.metadata.tag_systems.tag_base import MetaTagAll, MetaTagsBase
from weitsicht.utils import ResultFailure


class ImageFromMetaBuilder:
    """Build an :class:`~weitsicht.image.perspective.ImagePerspective` from metadata.

    The builder caches intermediate results (IOR/EOR) and provides a single place to add
    tracing/debug information in the future.
    """

    def __init__(self, tags: MetaTagsBase | MetaTagAll):
        self.tags: MetaTagAll = tags.get_all() if isinstance(tags, MetaTagsBase) else tags
        self._ior: IORFromMetaResult | None = None
        self._eor: EORFromMetaResult | None = None

    def ior(self) -> IORFromMetaResult:
        if self._ior is None:
            self._ior = ior_from_meta(tags_ior=self.tags.ior_base, tags_ior_extended=self.tags.ior_extended)
        return self._ior

    def eor(
        self,
        crs: CRS | None = None,
        vertical_ref: str = "ellipsoidal",
        height_rel: float = 0.0,
    ) -> EORFromMetaResult:
        if self._eor is None:
            self._eor = eor_from_meta(tags=self.tags, crs=crs, vertical_ref=vertical_ref, height_rel=height_rel)
        return self._eor

    def image(
        self,
        crs: CRS | None = None,
        vertical_ref: str = "ellipsoidal",
        height_rel: float = 0.0,
    ) -> ImageFromMetaResult:
        ior_res = self.ior()
        eor_res = self.eor(crs=crs, vertical_ref=vertical_ref, height_rel=height_rel)

        if ior_res.ok is False or eor_res.ok is False:
            errors: list[str] = []
            issues: set[MetadataIssue] = set()
            if ior_res.ok is False:
                errors.append(ior_res.error)
                issues.update(ior_res.issues)
            if eor_res.ok is False:
                errors.append(eor_res.error)
                issues.update(eor_res.issues)
            return ResultFailure(ok=False, error="; ".join(errors), issues=issues)

        image = ImagePerspective(
            width=ior_res.width,
            height=ior_res.height,
            camera=ior_res.camera,
            position=eor_res.position,
            orientation=eor_res.orientation,
            crs=eor_res.crs,
        )
        return ImageFromMetaResultSuccess(ok=True, image=image, ior=ior_res, eor=eor_res)


def image_from_meta(
    tags: MetaTagsBase | MetaTagAll,
    crs: CRS | None = None,
    vertical_ref: str = "ellipsoidal",
    height_rel: float = 0.0,
) -> ImageFromMetaResult:
    """Build an image (camera + EOR) from metadata.

    :param tags: Tag resolver instance or already-resolved tag container.
    :type tags: MetaTagsBase | MetaTagAll
    :param crs: Optional override CRS for interpreting GPS tags, defaults to ``None``.
    :type crs: CRS | None
    :param vertical_ref: Vertical reference mode passed to :func:`eor_from_meta`, defaults to ``ellipsoidal``.
    :type vertical_ref: str
    :param height_rel: Reference height (meters) for ``vertical_ref='relative'``, defaults to ``0.0``.
    :type height_rel: float
    :return: Successful image build result or a failure result.
    :rtype: ImageFromMetaResult
    """

    return ImageFromMetaBuilder(tags).image(crs=crs, vertical_ref=vertical_ref, height_rel=height_rel)



import numpy as np
import pytest

from weitsicht import Rotation


def test_from_opk_degree():

    r = Rotation.from_opk_degree(45, 0, 45)
    assert np.allclose(np.array([45, 0, 45]), r.opk_degree, atol=1e-9, rtol=0)
    assert np.allclose(np.array([np.pi / 4, 0, np.pi / 4]), r.opk, atol=1e-9, rtol=0)


def test_from_opk():
    r = Rotation.from_opk(np.pi / 4, 0, np.pi / 4)
    assert np.allclose(np.array([45, 0, 45]), r.opk_degree, atol=1e-9, rtol=0)
    assert np.allclose(np.array([np.pi / 4, 0, np.pi / 4]), r.opk, atol=1e-9, rtol=0)


def test_from_apk_degree():
    r = Rotation.from_apk_degree(30, 20, -10)
    assert np.allclose(np.array([30, 20, -10]), r.apk_degree, atol=1e-9, rtol=0)
    assert np.allclose(np.array([np.deg2rad(30), np.deg2rad(20), np.deg2rad(-10)]), r.apk, atol=1e-9, rtol=0)


def test_from_apk():
    r = Rotation.from_apk(np.deg2rad(30), np.deg2rad(20), np.deg2rad(-10))
    assert np.allclose(np.array([30, 20, -10]), r.apk_degree, atol=1e-9, rtol=0)
    assert np.allclose(np.array([np.deg2rad(30), np.deg2rad(20), np.deg2rad(-10)]), r.apk, atol=1e-9, rtol=0)


def test_matrix():
    mat_test = np.array(
        [
            [0.70710678, -0.70710678, 0.0],
            [0.5, 0.5, -0.70710678],
            [0.5, 0.5, 0.70710678],
        ]
    )

    r = Rotation.from_opk_degree(45, 0, 45)
    assert np.allclose(r.matrix, mat_test, atol=1e-6, rtol=0)

    r.matrix = mat_test

    assert np.allclose(np.array([45, 0, 45]), r.opk_degree, atol=1e-9, rtol=0)


def test_opk_degree():

    r = Rotation.from_opk_degree(0, 0, 0)

    assert np.allclose(np.eye(3, 3), r.matrix)

    with pytest.raises(ValueError):
        r.opk_degree = np.array([0, 0, 0, 0])

    assert np.allclose(np.array([0, 0, 0]), r.opk, atol=1e-9, rtol=0)

    r.opk_degree = np.array([45, 0, 45])
    assert np.allclose(np.array([45, 0, 45]), r.opk_degree, atol=1e-9, rtol=0)
    assert np.allclose(np.array([np.pi / 4, 0, np.pi / 4]), r.opk, atol=1e-9, rtol=0)


def test_opk():

    r = Rotation.from_opk_degree(0, 0, 0)
    r.opk = np.array([0, 0, 0])

    assert np.allclose(np.eye(3, 3), r.matrix, atol=1e-9, rtol=0)

    with pytest.raises(ValueError):
        r.opk = np.array([0, 0, 0, 0])

    assert np.allclose(np.array([0, 0, 0]), r.opk, atol=1e-9, rtol=0)


def test_apk_degree():
    r = Rotation.from_apk_degree(0, 0, 0)

    assert np.allclose(np.eye(3, 3), r.matrix)

    with pytest.raises(ValueError):
        r.apk_degree = np.array([0, 0, 0, 0])

    assert np.allclose(np.array([0, 0, 0]), r.apk, atol=1e-9, rtol=0)

    r.apk_degree = np.array([30, 20, -10])
    assert np.allclose(np.array([30, 20, -10]), r.apk_degree, atol=1e-9, rtol=0)
    assert np.allclose(np.array([np.deg2rad(30), np.deg2rad(20), np.deg2rad(-10)]), r.apk, atol=1e-9, rtol=0)


def test_apk():
    r = Rotation.from_apk_degree(0, 0, 0)
    r.apk = np.array([0, 0, 0])

    assert np.allclose(np.eye(3, 3), r.matrix, atol=1e-9, rtol=0)

    with pytest.raises(ValueError):
        r.apk = np.array([0, 0, 0, 0])

    assert np.allclose(np.array([0, 0, 0]), r.apk, atol=1e-9, rtol=0)


@pytest.mark.xfail(reason="Not implemented yet")
def test_transform_to_crs():

    assert True



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


import numpy as np
import pytest
from pyproj import CRS

from weitsicht import Issue
from weitsicht.mapping.horizontal_plane import MappingHorizontalPlane


def test_from_dict():
    mapper = MappingHorizontalPlane.from_dict({"type": "horizontalPlane", "plane_altitude": 43, "crs": "EPSG:4326+3855"})
    assert mapper is not None

    with pytest.raises(KeyError):
        MappingHorizontalPlane.from_dict({"type_key_missing": "whatsoever"})

    with pytest.raises(ValueError):
        MappingHorizontalPlane.from_dict({"type": "horizontalPlaneWrong"})


def test_param_dict():

    mapper = MappingHorizontalPlane.from_dict({"type": "horizontalPlane", "plane_altitude": 43.0, "crs": "EPSG:4979"})

    param = mapper.param_dict
    assert param == {
        "type": mapper.type.fullname,
        "plane_altitude": 43.0,
        "crs": mapper.crs.to_wkt() if mapper.crs is not None else None,
    }

    assert mapper.crs is not None
    assert mapper.crs.equals(CRS("EPSG:4979"))


def test_map_coordinates_from_rays():
    mapper = MappingHorizontalPlane.from_dict({"type": "horizontalPlane", "plane_altitude": 43.0, "crs": "EPSG:4979"})
    mapping_result = mapper.map_coordinates_from_rays(
        ray_start_crs_s=np.array([0, 0, 100]),
        ray_vectors_crs_s=np.array([0, 0, -1]),
        crs_s=CRS("EPSG:25833"),
    )
    assert mapping_result.ok is True
    assert np.linalg.norm(mapping_result.coordinates - np.array([0, 0, 43])) < 0.0000001

    # test wrong direction

    mapping_result = mapper.map_coordinates_from_rays(
        ray_start_crs_s=np.array([0, 0, 100]),
        ray_vectors_crs_s=np.array([0, 0, 1]),
        crs_s=CRS("EPSG:25833"),
    )
    assert mapping_result.ok is False

    # Two rays, one fails one works
    mapping_result = mapper.map_coordinates_from_rays(
        ray_start_crs_s=np.array([[0, 0, 100], [0, 0, 100]]),
        ray_vectors_crs_s=np.array([[0, 0, 1], [0, 0, -1]]),
        crs_s=CRS("EPSG:25833"),
    )
    assert mapping_result.ok is True
    assert bool(np.any(np.isnan(mapping_result.coordinates[0]))) is True
    assert bool(np.any(np.isnan(mapping_result.coordinates[1]))) is False
    assert Issue.NO_INTERSECTION in mapping_result.issues

    # test parallel line
    mapping_result = mapper.map_coordinates_from_rays(
        ray_start_crs_s=np.array([0, 0, 100]),
        ray_vectors_crs_s=np.array([0.3, 0.4, 0.0]),
        crs_s=CRS("EPSG:25833"),
    )
    assert mapping_result.ok is False

    # Test different sizes of start and ray vector
    with pytest.raises(ValueError):
        mapper.map_coordinates_from_rays(
            ray_start_crs_s=np.array([[0, 0, 150], [0, 0, 100]]),
            ray_vectors_crs_s=np.array([0, 0, -1]),
            crs_s=CRS("EPSG:25833"),
        )


def test_map_heights_from_coordinates():

    # Different CRS and 3d coordinates used
    mapper = MappingHorizontalPlane.from_dict({"type": "horizontalPlane", "plane_altitude": 43.0, "crs": "EPSG:4979"})

    coordinates = np.array([602013.0, 5340384.696, 400.0])
    mapping_result = mapper.map_heights_from_coordinates(coordinates_crs_s=coordinates, crs_s=CRS("EPSG:25833"))
    assert mapping_result.ok is True
    assert np.linalg.norm(mapping_result.coordinates - np.array([602013.0, 5340384.696, 43.0])) < 0.00001

    # Same CRS and only 2D coordinates are used as input
    mapper = MappingHorizontalPlane.from_dict({"type": "horizontalPlane", "plane_altitude": 43.0, "crs": "EPSG:25833"})

    coordinates = np.array([602013.0, 5340384.696])
    mapping_result = mapper.map_heights_from_coordinates(coordinates_crs_s=coordinates, crs_s=CRS("EPSG:25833"))

    assert mapping_result.ok is True
    assert np.linalg.norm(mapping_result.coordinates - np.array([602013.0, 5340384.696, 43.0])) < 0.00001



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Coordinate transformation helpers based on :mod:`pyproj`."""

from __future__ import annotations

import logging
from functools import lru_cache
from typing import overload

import numpy as np
from pyproj import CRS, Transformer
from pyproj.exceptions import CRSError, ProjError
from pyproj.transformer import TransformerGroup
from shapely import errors as shapely_errors
from shapely import geometry

from weitsicht import cfg
from weitsicht.exceptions import CoordinateTransformationError
from weitsicht.utils import ArrayNx2, ArrayNx3, Vector2D, Vector3D

__all__ = ["CoordinateTransformer"]

logger = logging.getLogger(__name__)

# The doc states that cached Versions can be used to speed up performance significantly
TransformerFromCRS = lru_cache(Transformer.from_crs)
TransformerFromPipeline = lru_cache(Transformer.from_pipeline)


def get_transformation_transformergroup(crs_source: CRS, crs_target: CRS) -> TransformerGroup:  # pragma: no cover
    """Create a :class:`pyproj.transformer.TransformerGroup` between two CRSs.

    Note: the pyproj documentation states that the returned objects are not thread-safe.

    :param crs_source: Source CRS.
    :type crs_source: CRS
    :param crs_target: Target CRS.
    :type crs_target: CRS
    :return: Transformer group for ``crs_source`` → ``crs_target``.
    :rtype: TransformerGroup
    :raises CoordinateTransformationError: If no suitable transformation can be established.
    """

    try:
        transformer = TransformerGroup(
            crs_source,
            crs_target,
            always_xy=True,
            allow_ballpark=cfg._ballpark_transformation,
        )
        if cfg._only_best_transformation and not transformer.best_available:
            raise CoordinateTransformationError(
                "Best Transformation is not available but only best transformations are allowed"
            )
        return transformer

        # self.transformer = Transformer.from_crs(self._crs, crs_target, always_xy=True,
        #                                        only_best=cfg._only_best_transformation,
        #                                        allow_ballpark=cfg._ballpark_transformation)
    except ProjError as e:
        raise CoordinateTransformationError("Transformation could not be established.") from e


class CoordinateTransformer:
    """Class to handle transformations of coordinates.
    Basically a little pyproj wrapper with basic functions needed in the package.
    """

    @overload
    def __init__(self, transformer: Transformer): ...
    @overload
    def __init__(self, transformer: list[Transformer]): ...
    def __init__(self, transformer: Transformer | list[Transformer]):
        """Create a transformer wrapper.

        :param transformer: A single pyproj transformer or a chain of transformers.
        :type transformer: Transformer | list[Transformer]
        """

        if isinstance(transformer, Transformer):
            self._transformers = [transformer]
        else:
            self._transformers: list[Transformer] = transformer

    @property
    def source_crs(self) -> CRS | None:
        """Return the source CRS of the transformer chain."""
        return self._transformers[0].source_crs

    @property
    def target_crs(self) -> CRS | None:
        """Return the target CRS of the transformer chain."""
        return self._transformers[-1].target_crs

    @classmethod
    def from_pipeline(cls, proj_pipeline: str):
        """Create a transformer from a PROJ pipeline definition.

        :param proj_pipeline: PROJ pipeline string.
        :type proj_pipeline: str
        :return: Transformer instance.
        :rtype: CoordinateTransformer
        :raises CoordinateTransformationError: If the pipeline cannot be established.
        """

        try:
            transformer = TransformerFromPipeline(proj_pipeline)

        except (CRSError, ProjError) as e:
            raise CoordinateTransformationError("Establishing PROJ PIPELINE not working") from e

        _obj = cls(transformer=[transformer])
        return _obj

    @classmethod
    def from_crs(cls, crs_s: CRS | None, crs_t: CRS | None) -> CoordinateTransformer | None:
        """Create a coordinate transformer from source CRS to target CRS.

        If both CRSs are ``None`` or equal, this returns ``None`` (no transformation required).

        :param crs_s: Source CRS, or ``None``.
        :type crs_s: CRS | None
        :param crs_t: Target CRS, or ``None``.
        :type crs_t: CRS | None
        :return: Coordinate transformer or ``None`` if no transformation is required.
        :rtype: CoordinateTransformer | None
        :raises ValueError: If only one CRS is ``None``.
        :raises CoordinateTransformationError: If establishing a transformation fails.
        """

        # First we check if we anyhow need any transformation
        # If both CRS are None or both are equal we will not do any transformation
        if crs_s is None and crs_t is None:
            return None

        if (crs_s is None) != (crs_t is None):
            raise ValueError("Either both CRS have to be stated or for both None")

        assert crs_s is not None
        assert crs_t is not None
        if crs_s.equals(crs_t):
            return None

        transformer: list[Transformer] = []

        geod_s = crs_s.geodetic_crs.to_3d()
        geod_t = crs_t.geodetic_crs.to_3d()

        if geod_s.equals(geod_t):
            transformer.append(
                TransformerFromCRS(
                    crs_s,
                    crs_t,
                    always_xy=True,
                    allow_ballpark=cfg._ballpark_transformation,
                    only_best=cfg._only_best_transformation,
                )
            )

        else:
            try:
                transformer.append(
                    TransformerFromCRS(
                        crs_s,
                        geod_s,
                        always_xy=True,
                        allow_ballpark=cfg._ballpark_transformation,
                        only_best=cfg._only_best_transformation,
                    )
                )

                transformer.append(
                    TransformerFromCRS(
                        geod_s,
                        geod_t,
                        always_xy=True,
                        allow_ballpark=cfg._ballpark_transformation,
                        only_best=cfg._only_best_transformation,
                    )
                )

                transformer.append(
                    TransformerFromCRS(
                        geod_t,
                        crs_t,
                        always_xy=True,
                        allow_ballpark=cfg._ballpark_transformation,
                        only_best=cfg._only_best_transformation,
                    )
                )
            except (CRSError, ProjError) as e:
                raise CoordinateTransformationError("Establishing coordinate transformation for mapping failed") from e

        _obj = cls(transformer=transformer)

        return _obj

    @staticmethod
    def is_3d(coo: ArrayNx2 | ArrayNx3):
        """Return whether the coordinate array has a Z component.

        :param coo: Coordinate array.
        :type coo: ArrayNx2 | ArrayNx3
        :return: ``True`` if the input has shape N×3.
        :rtype: bool
        """
        if coo.shape[1] == 3:
            return True
        return False

    def transform(self, coordinates: ArrayNx3 | ArrayNx2, direction: str = "forward") -> ArrayNx3 | ArrayNx2:
        """Transform coordinates using the configured transformer chain.

        :param coordinates: Coordinates to transform as N×2 (2D) or N×3 (3D).
        :type coordinates: ArrayNx3 | ArrayNx2
        :param direction: Transform direction (``forward``, ``inverse``, ``identity``), defaults to ``forward``.
        :type direction: str
        :return: Transformed coordinates with the same dimensionality as input.
        :rtype: ArrayNx3 | ArrayNx2
        :raises CoordinateTransformationError: If input dimensions are invalid or a transformation fails.
        """

        _coordinates = coordinates.copy()
        if _coordinates.ndim == 1:
            _coordinates = np.array([_coordinates])

        if _coordinates.shape[1] < 2:
            raise CoordinateTransformationError("Wrong input dimension")

        # Iterate over the transformer list
        # if direction is inverse we need to start from end for list
        if direction == "inverse":
            transformers_order = list(reversed(self._transformers))
        else:
            transformers_order = self._transformers

        for trafo in transformers_order:
            try:
                # Other ways I called transform
                # res = trafo.transform(*_coordinates.T, errcheck=True)
                # p_crs = np.array(res).T
                # fx, fy, fz = transformer_srs_4979.transform(ray_start[:, 0], ray_start[:, 1],
                #                                            ray_start[:, 2], errcheck=True)

                if self.is_3d(_coordinates):
                    res = trafo.transform(
                        _coordinates[:, 0],
                        _coordinates[:, 1],
                        _coordinates[:, 2],
                        errcheck=True,
                        direction=direction,
                    )
                    _coordinates = np.array(res).T
                else:
                    res = trafo.transform(
                        _coordinates[:, 0],
                        _coordinates[:, 1],
                        errcheck=True,
                        direction=direction,
                    )
                    _coordinates = np.array(res).T

                if np.any(_coordinates == np.inf):
                    raise CoordinateTransformationError("Coordinate Transformation failed")

            except (CRSError, ProjError) as e:
                raise CoordinateTransformationError("Error while transforming coordinates") from e

        return _coordinates


class Geometries:  # pragma: no cover
    pass

    def __init__(self, points: ArrayNx3 | Vector3D | Vector2D | ArrayNx2):
        # points: ArrayNx3 | Vector3D | Vector2D | ArrayNx2):
        self._coordinates = points.copy()

        # The class uses internal only array with ndim 2. So point np.array([0,0,1]) will be np.array([[0,0,1]])
        if self._coordinates.ndim == 1:
            self._coordinates = np.array([self._coordinates])

    @property
    def is_point(self) -> bool:
        if self._coordinates.shape[0] == 1:
            return True
        return False

    @property
    def is_3d(self):
        if self._coordinates.shape[1] == 3:
            return True
        return False

    def geojson(self, geom_type: str) -> dict:
        """Convert the stored coordinates into a GeoJSON mapping.

        :param geom_type: Geometry type (``Point``, ``LineString``, ``Polygon``).
        :type geom_type: str
        :return: GeoJSON mapping dict.
        :rtype: dict
        :raises CoordinateTransformationError: If geometry type is unsupported or coordinates are invalid.
        """
        try:
            if geom_type == "Point":
                geojson = geometry.Point(self.coordinates)
            elif geom_type == "LineString":
                geojson = geometry.LineString(self._coordinates)
            elif geom_type == "Polygon":
                geojson = geometry.Polygon(self._coordinates)
            else:
                raise CoordinateTransformationError(r"Geometry is not supported. Only 'Point', 'LineString', 'Polygon'")

        except (shapely_errors.GEOSException, ValueError) as err:
            # Will also raise if multi coordinates and point type is used
            raise CoordinateTransformationError("GeoJSON can not be established from given coordinates") from err

        geojson = geometry.mapping(geojson)

        return geojson

    @property
    def coordinates(self) -> ArrayNx3 | ArrayNx2:
        return self._coordinates

    @coordinates.setter
    def coordinates(self, new_coordinate: ArrayNx3 | ArrayNx2):

        points = new_coordinate.copy()

        # The class uses internal only array with ndim 2. So point np.array([0,0,1]) will be np.array([[0,0,1]])
        if points.ndim == 1:
            new_coordinate = np.array([new_coordinate])
        self._coordinates = new_coordinate



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Mapping backend for a horizontal plane at a fixed altitude."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any

import numpy as np
import pyproj.exceptions
from pyproj import CRS

from weitsicht.exceptions import CRSInputError
from weitsicht.geometry.intersection_plane import intersection_plane_mat_operation
from weitsicht.mapping.base_class import MappingBase, MappingType
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import (
    ArrayNx2,
    ArrayNx3,
    Issue,
    MappingResult,
    MappingResultSuccess,
    ResultFailure,
    to_array_nx3,
)

__all__ = ["MappingHorizontalPlane"]

logger = logging.getLogger(__name__)


class MappingHorizontalPlane(MappingBase):
    """Mapping backend for a horizontal plane at a fixed altitude.

    The plane is defined by ``plane_altitude`` (Z coordinate in mapper CRS) and a fixed normal (0, 0, 1).
    If a CRS transformation is performed, the plane altitude is transformed appropriately via the provided
    ``crs_s``/``transformer`` inputs.
    """

    def __init__(self, plane_altitude: float | int = 0.0, crs: CRS | None = None):
        """Create a horizontal plane mapper.

        :param plane_altitude: Altitude of the horizontal plane (Z value in mapper CRS), defaults to ``0.0``.
        :type plane_altitude: float | int
        :param crs: CRS of the plane, defaults to ``None``.
        :type crs: CRS | None
        """

        super().__init__()

        self.plane_altitude = float(plane_altitude)
        self.plane_normal = np.array([0, 0, 1])
        self._crs = crs

        # TODO check if crs is not None that the CRS provides a vertical reference

    @classmethod
    def from_dict(cls, mapper_dict: Mapping[str, Any]) -> MappingHorizontalPlane:
        """Create a :class:`MappingHorizontalPlane` instance from a configuration dictionary.

        Required keys:
        - ``type``: must be ``horizontalPlane``

        Optional keys:
        - ``plane_altitude``: plane altitude (float), defaults to ``0.0``
        - ``crs``: CRS WKT string, defaults to ``None``

        :param mapper_dict: Mapper configuration dictionary (typically created via ``mapper.param_dict``).
        :type mapper_dict: Mapping[str, Any]
        :return: Instantiated mapper.
        :rtype: MappingHorizontalPlane
        :raises KeyError: If the dictionary key ``type`` is missing.
        :raises ValueError: If the mapper type is unsupported.
        :raises CRSInputError: If the CRS WKT string is invalid.
        """

        if mapper_dict.get("type", None) is None:
            raise KeyError("Dictionary key 'type' is missing")

        if mapper_dict.get("type", None) != MappingType.HorizontalPlane.fullname:
            raise ValueError("Mapper dictionary type is not " + MappingType.HorizontalPlane.fullname)

        plane_altitude = mapper_dict.get("plane_altitude", 0.0)
        crs_text = mapper_dict.get("crs", None)

        crs = None
        if crs_text is not None:
            try:
                crs = CRS(crs_text)
            except pyproj.exceptions.CRSError as err:
                raise CRSInputError("No valid CRS WKT string supported") from err

        mapping = cls(plane_altitude=plane_altitude, crs=crs)
        return mapping

    @property
    def type(self):
        """Return the mapper type."""
        return MappingType.HorizontalPlane

    @property
    def param_dict(self):
        """Return the mapper configuration dictionary.

        :return: Mapper configuration dictionary.
        :rtype: dict[str, Any]
        """

        crs_text = None
        if self._crs is not None:
            crs_text = self.crs_wkt

        return {
            "type": self.type.fullname,
            "plane_altitude": self.plane_altitude,
            "crs": crs_text,
        }

    def map_coordinates_from_rays(
        self,
        ray_vectors_crs_s: ArrayNx3,
        ray_start_crs_s: ArrayNx3,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map coordinates from rays to a horizontal plane.

        :param ray_vectors_crs_s: Ray direction vectors (N×3).
        :type ray_vectors_crs_s: ArrayNx3
        :param ray_start_crs_s: Ray start points (N×3), same shape as ``ray_vectors_crs_s``.
        :type ray_start_crs_s: ArrayNx3
        :param crs_s: CRS of the input rays, defaults to None.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to None.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing intersection coordinates and a validity mask.
        :rtype: MappingResult
        :raises ValueError: If input arrays have incompatible shapes.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        # A Array(3,) (Vector3D) should also work ray_vectors_crs_s and ray_start_crs_s
        ray_pos = to_array_nx3(ray_start_crs_s)
        ray_vec = to_array_nx3(ray_vectors_crs_s)

        if ray_pos.shape != ray_vec.shape:
            raise ValueError("Ray and Pos Vector not the same size")

        # transform the given coordinates to the mappers crs
        # replace the z coordinate of the coordinates in mapper crs with the mappers height
        # transform back to the source CRS which now provides mapper height in source crs
        # Therefore the mapper's height is correct transformed to the image crs at the image's prj center
        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        # We will do this iterative, so if in the coordinate transformation raster are used
        # We will iterate the plane height to that one from the correct intersection point
        # Thus also valid mask should be taken into account, otherwise coordinate operation are wrong
        # If both coo_trafo is None we will skip iteration

        prev_intersect_points = np.ones(ray_pos.shape, dtype=float)
        diff_greater_1mm = np.ones(ray_pos.shape, dtype=float)
        iteration: int = 0
        valid_mask = np.ones((ray_pos.shape[0],), dtype=bool)
        plane_point_source_crs = np.ones(ray_pos.shape, dtype=float)
        intersect_points = np.empty(ray_pos.shape, dtype=float)
        intersect_points.fill(np.nan)

        # initial height will be taken from ray start position
        if coo_trafo is not None:
            plane_point_mapper_crs = coo_trafo.transform(ray_pos)
        else:
            plane_point_mapper_crs = ray_pos * 1.0

        plane_point_mapper_crs[:, 2] = self.plane_altitude

        while np.any(diff_greater_1mm > 0.001) and iteration < 20:
            index_valid_mask = np.flatnonzero(valid_mask)
            # Transform back to source crs. So we have the horizontal planes Altitude in the correct system
            if coo_trafo is not None:
                plane_point_source_crs[index_valid_mask, :] = coo_trafo.transform(
                    plane_point_mapper_crs[index_valid_mask, :], direction="inverse"
                )
            else:
                plane_point_source_crs[index_valid_mask, :] = plane_point_mapper_crs[index_valid_mask, :] * 1.0

            # _intersect_points, _valid_mask are corresponding to size of true items in valid_mask
            # its very unlikely that validity changes between iterations
            # only if the plane point would then be exactly the ray pos it would be invalid
            # so at least always grapping validity we can reduce risk here
            # Testing that case would be lot of work, creating a pipeline which transform exactly like this.
            _intersect_points, _valid_mask = intersection_plane_mat_operation(
                ray_vec[index_valid_mask, :],
                ray_pos[index_valid_mask, :],
                plane_point=plane_point_source_crs[index_valid_mask, :],
                plane_normal=self.plane_normal,
            )
            _index_false = np.flatnonzero(~_valid_mask)
            _index_correct = np.flatnonzero(_valid_mask)
            valid_mask[index_valid_mask[_index_false]] = False
            intersect_points[index_valid_mask[_index_correct], :] = _intersect_points[_index_correct, :]
            if coo_trafo is None:
                break
            if not np.any(valid_mask):
                break
            diff_greater_1mm = np.abs(intersect_points[valid_mask, :] - prev_intersect_points[valid_mask, :])
            prev_intersect_points = intersect_points * 1.0
            iteration += 1
            # Now the new plane point is the previous intersection point

            if coo_trafo is not None:
                plane_point_mapper_crs[valid_mask, :] = coo_trafo.transform(intersect_points[valid_mask, :])
                plane_point_mapper_crs[valid_mask, 2] = self.plane_altitude
            else:
                plane_point_mapper_crs[valid_mask, :] = ray_pos[valid_mask, :] * 1.0

        issue = set()
        if not np.all(valid_mask):
            issue = {Issue.NO_INTERSECTION}
            if not np.any(valid_mask):
                return ResultFailure(
                    ok=False,
                    error="None of the rays intersects",
                    issues={Issue.NO_INTERSECTION},
                )

        return MappingResultSuccess(
            ok=True,
            coordinates=intersect_points,
            mask=valid_mask,
            crs=crs_s,
            issues=issue,
        )

    def map_heights_from_coordinates(
        self,
        coordinates_crs_s: ArrayNx3 | ArrayNx2,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Assign the plane altitude as height for given coordinates.

        :param coordinates_crs_s: Coordinates to sample (N×2 or N×3).
        :type coordinates_crs_s: ArrayNx3 | ArrayNx2
        :param crs_s: CRS of the input coordinates, defaults to None.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to None.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing coordinates with plane altitude and a validity mask.
        :rtype: MappingResult
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        # Prepare input if something else than ArrayNx3 (e.g. List, ArrayN)
        coordinates = to_array_nx3(coordinates_crs_s, 0.0)

        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        if coo_trafo is not None:
            coo_mapper_crs = coo_trafo.transform(coordinates)
        else:
            coo_mapper_crs = coordinates * 1.0

        # Now change the altitude to the plane height
        coo_mapper_crs[:, 2] = self.plane_altitude

        # Transform back to source crs. So we have the horizontal planes Altitude in the correct system
        # Here no iteration is needed.
        # Or at least if we not assume that the transformation with a new height will not change the 2d coordinates
        if coo_trafo is not None:
            coo_source_crs = coo_trafo.transform(coo_mapper_crs, direction="inverse")
        else:
            coo_source_crs = coo_mapper_crs * 1.0

        # In that case all are valid because its simple assigned its heights
        # Any coordinate error would have already raised an error before
        mask = np.ones((coo_source_crs.shape[0]), dtype=bool)
        return MappingResultSuccess(ok=True, coordinates=coo_source_crs, mask=mask, crs=crs_s)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Base interfaces for mapping backends.

Mapping backends implement the conversion between rays/coordinates and mapped 3D coordinates
in a target domain (e.g. a horizontal plane, a raster DEM, a mesh).
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping
from enum import Enum
from typing import Any, TypeVar

from pyproj import CRS

from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import ArrayNx2, ArrayNx3, MappingResult

__all__ = ["MappingType", "MappingBase"]


class MappingType(Enum):
    """Enum identifying available mapping backends.

    The :attr:`fullname` attribute is a stable, human-readable identifier used for
    (de-)serialization (e.g. in :meth:`~weitsicht.mapping.base_class.MappingBase.param_dict`).
    """

    fullname: str  # human-readable identifier set on each enum member
    HorizontalPlane = 1, "horizontalPlane"
    Raster = 2, "Raster"
    GeoreferencedNumpyArray = 2, "GeorefArray"
    Trimesh = 3, "Trimesh"

    def __new__(cls, value: int, name: str):
        member = object.__new__(cls)
        member._value_ = value
        member.fullname = name
        return member

    def __int__(self):
        return self.value


_Self = TypeVar("_Self", bound="MappingBase")


class MappingBase:
    """Abstract base class for all mapping backends."""

    def __init__(self):
        """Initialize a mapper with an optional CRS (set via :attr:`crs`)."""
        self._crs: CRS | None = None

    @classmethod
    @abstractmethod
    def from_dict(cls: type[_Self], mapper_dict: Mapping[str, Any]) -> _Self:
        """Create a mapper instance from a configuration dictionary.

        :param mapper_dict: Mapper configuration dictionary (typically created via ``mapper.param_dict``).
        :type mapper_dict: Mapping[str, Any]
        :return: Instantiated mapper.
        :rtype: MappingBase
        :raises KeyError: If a required dictionary key is missing.
        :raises ValueError: If configuration values are invalid.
        :raises CRSInputError: If CRS input in the configuration is invalid (e.g. malformed WKT).
        :raises CRSnoZaxisError: If a required CRS does not define a Z axis.
        :raises MappingError: If the mapper cannot be initialized.
        """
        pass

    @property
    @abstractmethod
    def type(self) -> MappingType:
        """Return the mapper type.

        :return: Mapper type.
        :rtype: MappingType
        """
        pass

    @property
    @abstractmethod
    def param_dict(self) -> Mapping[str, Any]:
        """Return the mapper configuration dictionary.

        This dictionary is intended to be used with :meth:`from_dict`.

        :return: Mapper configuration dictionary.
        :rtype: Mapping[str, Any]
        """
        pass

    @property
    def crs(self) -> CRS | None:
        """Return the CRS of the mapper (if set).

        :return: CRS of the mapper.
        :rtype: CRS | None
        """
        return self._crs

    @crs.setter
    def crs(self, crs: CRS | None):
        """Set the CRS of the mapper.

        :param crs: CRS to use, or ``None`` to unset.
        :type crs: CRS | None
        """
        self._crs = crs

    @property
    def crs_wkt(self) -> str | None:
        """Return the CRS as WKT string, or ``None`` if not set.

        :return: CRS WKT string.
        :rtype: str | None
        """
        if self._crs is not None:
            return self._crs.to_wkt()
        return None

    @abstractmethod
    def map_coordinates_from_rays(
        self,
        ray_vectors_crs_s: ArrayNx3,
        ray_start_crs_s: ArrayNx3,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map 3D coordinates from ray definitions.

        The input rays are provided by a ray start point and a ray direction vector. If a CRS
        conversion is required, pass either ``crs_s`` (and the mapper's :attr:`crs` must be set)
        or a preconfigured ``transformer``.

        :param ray_vectors_crs_s: Ray direction vectors (N×3).
        :type ray_vectors_crs_s: ArrayNx3
        :param ray_start_crs_s: Ray start points (N×3), same shape as ``ray_vectors_crs_s``.
        :type ray_start_crs_s: ArrayNx3
        :param crs_s: CRS of the input rays, defaults to None.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to None.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing mapped coordinates and a validity mask.
        :rtype: MappingResult
        :raises ValueError: If input arrays have incompatible shapes.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """
        pass

    @abstractmethod
    def map_heights_from_coordinates(
        self,
        coordinates_crs_s: ArrayNx3 | ArrayNx2,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Sample heights for given 2D/3D coordinates.

        This is typically used to lift planar coordinates (x, y) to 3D (x, y, z) using a DEM.

        :param coordinates_crs_s: Coordinates to sample (N×2 or N×3). If N×2, Z is backend-defined.
        :type coordinates_crs_s: ArrayNx3 | ArrayNx2
        :param crs_s: CRS of the input coordinates, defaults to None.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to None.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing coordinates with sampled heights and a validity mask.
        :rtype: MappingResult
        :raises ValueError: If input arrays have incompatible shapes.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        pass



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Factory helpers for creating mapping backends from configuration dictionaries."""

from __future__ import annotations

from weitsicht.mapping.base_class import MappingBase, MappingType
from weitsicht.mapping.horizontal_plane import MappingHorizontalPlane
from weitsicht.mapping.map_trimesh import MappingTrimesh
from weitsicht.mapping.raster import MappingRaster

__all__ = ["get_mapper_from_dict"]


def get_mapper_from_dict(
    mapper_config: dict,
) -> MappingBase:
    """Load a mapper instance from a configuration dictionary.

    The configuration must include a ``type`` key matching an implemented mapper type
    (e.g. ``horizontalPlane``, ``Raster``, ``Trimesh``).

    Exceptions raised by the underlying mapper ``from_dict`` implementations are propagated unchanged.

    :param mapper_config: Mapper configuration dictionary (typically created via ``mapper.param_dict``).
    :type mapper_config: dict
    :return: Instantiated mapper.
    :rtype: MappingBase
    :raises KeyError: If the dictionary key ``type`` is missing.
    :raises FileNotFoundError: If a referenced raster/mesh file does not exist.
    :raises ValueError: If configuration values are invalid or the mapper type is unsupported.
    :raises CRSInputError: If CRS input in the configuration is invalid (e.g. malformed WKT).
    :raises CRSnoZaxisError: If a required CRS does not define a Z axis.
    :raises MappingError: If the mapper cannot be initialized.
    """

    mapper_type = mapper_config.get("type", None)
    if mapper_type is None:
        raise KeyError("Dictionary key 'type' is missing")

    if mapper_type == MappingType.HorizontalPlane.fullname:
        mapper = MappingHorizontalPlane.from_dict(mapper_config)

    elif mapper_type == MappingType.Raster.fullname:
        mapper = MappingRaster.from_dict(mapper_config)

    elif mapper_type == MappingType.Trimesh.fullname:
        mapper = MappingTrimesh.from_dict(mapper_config)

    else:
        raise ValueError(f"Unsupported mapper type {mapper_type!r}")

    return mapper



from pathlib import Path

import pytest

from weitsicht.camera.camera_dict_selector import get_camera_from_dict

DATA_DIR = Path(__file__).parent.resolve() / "data"


def test_select_camera_from_dict():

    cam_param_dict = {
        "type": "OpenCV",
        "calib_width": 8256,
        "calib_height": 5504,
        "fx": 4678.567517082265,
        "fy": 4678.567517082265,
        "cx": 4199.726251723716,
        "cy": 2749.841800528425,
        "k1": -0.043081934066152315,
        "k2": 0.013751556689262834,
        "k3": 0.0,
        "k4": 0.0,
        "p1": -0.0007525372490219161,
        "p2": 0.0027883695798851496,
    }

    cam = get_camera_from_dict(cam_param_dict)

    assert cam.type.fullname == "OpenCV"

    # wrong dict
    wrong_dict = {}
    with pytest.raises(KeyError):
        get_camera_from_dict(wrong_dict)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import logging

from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.camera.camera_dict_selector import get_camera_from_dict
from weitsicht.camera.camera_types import CameraType
from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective

__all__ = ["CameraType", "CameraBasePerspective", "CameraType", "CameraOpenCVPerspective", "get_camera_from_dict"]

logger = logging.getLogger(__name__)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Camera type definitions used across the weitsicht camera module."""

from enum import Enum

## Moved out of base_perspective if other cameras are not of type BasePerspective

__all__ = ["CameraType"]


class CameraType(Enum):
    """Enum of supported camera model types.

    The ``fullname`` attribute is used for serialization (e.g. in ``param_dict``)
    and must match what factory functions expect (e.g. ``get_camera_from_dict``).
    Add new camera model types here.
    """

    fullname: str
    Base = 1, "Base"
    OpenCV = 2, "OpenCV"

    def __new__(cls, value, name):
        member = object.__new__(cls)
        member._value_ = value
        member.fullname = name
        return member

    def __int__(self):
        return self.value



import json
from pathlib import Path
from typing import cast

from weitsicht import CameraOpenCVPerspective, CameraType, PyExifToolTags, ior_from_meta

DATA_DIR = Path(__file__).parent.resolve() / "data"


def test_estimate_camera_from_meta():

    meta_tag_dict = json.load(open(DATA_DIR / "image_2_exiftool_tags.json"))

    tags = PyExifToolTags(meta_data=meta_tag_dict)
    result = ior_from_meta(tags_ior=tags.get_ior_base(), tags_ior_extended=tags.get_ior_extended())

    assert result.ok is True

    assert (result.width, result.height) == (11648, 8736)
    assert result.camera is not None

    # This dict has DJI calibration info
    meta_tag_dict = json.load(open(DATA_DIR / "image_3_exiftool_tags.json"))
    tags = PyExifToolTags(meta_data=meta_tag_dict)
    result = ior_from_meta(tags_ior=tags.get_ior_base(), tags_ior_extended=tags.get_ior_extended())

    assert result.ok is True

    assert result.camera.type == CameraType.OpenCV
    assert isinstance(result.camera, CameraOpenCVPerspective)
    camera = cast(CameraOpenCVPerspective, result.camera)
    assert (result.width, result.height) == (5472, 3648)
    assert camera is not None
    assert camera.focal_length_for_gsd_in_pixel == 3666.666504
    assert camera.cy == 1824.0



from pathlib import Path

import pytest
from pyproj import CRS

from weitsicht.mapping.base_class import MappingType
from weitsicht.mapping.mapping_dict_selector import get_mapper_from_dict

DATA_DIR = Path(__file__).parent.resolve() / "data"


def test_mapper_load_from_dict():
    # wrong dictionary, no Type present
    dict_fail = {"raster_filepath": "wrongfile", "crs": CRS("EPSG:31287+5778").to_wkt()}

    with pytest.raises(KeyError):
        get_mapper_from_dict(dict_fail)

    # wrong type
    dict_fail = {
        "type": "PointCloud",
        "raster_filepath": "wrongfile",
        "crs": CRS("EPSG:31287+5778").to_wkt(),
    }

    with pytest.raises(ValueError):
        get_mapper_from_dict(dict_fail)

    # Horizontal dict working
    dict_horizontal = {
        "type": "horizontalPlane",
        "plane_altitude": 43.0,
        "crs": "EPSG:4979",
    }
    mapper = get_mapper_from_dict(dict_horizontal)
    assert mapper.type.fullname is MappingType.HorizontalPlane.fullname

    # Raster dict working
    dict_raster = {
        "type": "Raster",
        "raster_filepath": (DATA_DIR / "dhm_at_lamb_10m_2018.tif").as_posix(),
        "crs": CRS("EPSG:31287+5778").to_wkt(),
    }

    mapper = get_mapper_from_dict(dict_raster)
    assert mapper.type.fullname is MappingType.Raster.fullname

    # Raster Fail
    dict_raster = {
        "type": "Raster",
        "raster_filepath": "wrongfile",
        "crs": CRS("EPSG:31287+5778").to_wkt(),
    }

    with pytest.raises(FileNotFoundError):
        get_mapper_from_dict(dict_raster)

    # Mesh working dict
    mesh_path = DATA_DIR / "decimated_160k_31256.ply"
    dict_mesh = {
        "type": MappingType.Trimesh.fullname,
        "mesh_filepath": mesh_path.as_posix(),
        "crs": None,
    }
    mapper = get_mapper_from_dict(dict_mesh)
    assert mapper.type.fullname == MappingType.Trimesh.fullname

    # Mesh dict not working
    mesh_path = DATA_DIR / "decimated_160k_31256_wrong.ply"
    dict_mesh = {
        "type": MappingType.Trimesh.fullname,
        "mesh_filepath": mesh_path.as_posix(),
        "crs": None,
    }
    with pytest.raises(FileNotFoundError):
        get_mapper_from_dict(dict_mesh)
